import 'dart:ffi' as ffi;
import 'package:ffi/ffi.dart';
import 'package:msgpack_dart/msgpack_dart.dart';
import 'functions.dart';
import 'mdb.dart';

const int OP_MDB_FIRST = 0;
const int OP_MDB_NEXT = 8;
const int DENIED = 13;
const int MDB_CREATE = 0x40000;
const int MDB_RDONLY = 0x20000;
const int MDB_NOSUBDIR = 0x4000;
const int MDB_NOLOCK = 0x400000;
const int KEY_NOT_EXISTS = 4294936498;

class ErrorHandler {
  final int code;
  ErrorHandler(this.code) {
    if (code != 0) {
      print("error code number $code");
      final errorMessage = Utf8.fromUtf8(LMDBFunctions().mdbStrError(code));
      print("error message: $errorMessage");
      throw errorMessage;
    }
  }
}

class LMDB {
  static LMDB open(String path, {int maxDb = 1}) {
    LMDB box = _boxes[path];

    if (box == null) {
      final functions = LMDBFunctions();

      final pointerToMdbEnv = allocate<ffi.Pointer<MDBenv>>();

      /// Try to create the LMDB, if the access was denied because the database.lock
      /// file cannot be accessed by the current user, try with [MDB_NOLOCK] flag
      if (_createLMDB(functions, pointerToMdbEnv, path, maxDb,
              MDB_CREATE | MDB_NOSUBDIR) ==
          DENIED) {
        functions.mdbEnvClose(pointerToMdbEnv.value);

        /// Trying open ignoring flags
        ErrorHandler(_createLMDB(
          functions,
          pointerToMdbEnv,
          path,
          maxDb,
          MDB_CREATE | MDB_NOLOCK | MDB_NOSUBDIR,
        ));
      }

      box = LMDB._(path, pointerToMdbEnv.value);
      _boxes[path] = box;
    }

    return box;
  }

  static final _boxes = Map<String, LMDB>();

  final String _path;
  bool _isClosed = false;
  final ffi.Pointer<MDBenv> _mdbEnv;
  ffi.Pointer<ffi.IntPtr> _mdbDbi;

  LMDB._(this._path, this._mdbEnv);

  static int _createLMDB(
    final LMDBFunctions functions,
    final ffi.Pointer<ffi.Pointer<MDBenv>> pointerToMdbEnv,
    final String path,
    final int maxDatabases,
    final int flags,
  ) {
    var result = functions.mdbEnvCreate(pointerToMdbEnv);

    if (result != 0) return result;

    final ffi.Pointer<MDBenv> mdbEnv = pointerToMdbEnv.value;

    result = functions.mdbEnvSetMaxdbs(mdbEnv, maxDatabases);

    if (result != 0) return result;

    result = functions.mdbEnvOpen(
      mdbEnv,
      Utf8.toUtf8(path),
      flags,
      0664,
    );

    return result;
  }

  T find<T>(String key) {
    if (isClosed) {
      return null;
    }

    T result;

    _execute((functions, transaction, database) {
      final valuePointer = allocate<MDBvalue>();

      final response = functions.mdbGet(
        transaction,
        database.value,
        MDBvalue.fromUint8List(serialize(key)).addressOf,
        valuePointer,
      );
      if (response == KEY_NOT_EXISTS) {
        print('Key "$key" not registred');
        return;
      }
      ErrorHandler(response);
      result = _localize(valuePointer);
    }, isReadOnly: true);

    return result;
  }

  void write(String key, final Object value) {
    if (isClosed) {
      return null;
    }

    _execute(
      (functions, transaction, database) {
        ErrorHandler(
          functions.mdbPut(
            transaction,
            database.value,
            MDBvalue.fromUint8List(serialize(key)).addressOf,
            MDBvalue.fromUint8List(serialize(value)).addressOf,
            0,
          ),
        );
      },
    );
  }

  void delete(String key) {
    if (isClosed) {
      return null;
    }

    _execute(
      (functions, transaction, database) {
        final valuePointer = allocate<MDBvalue>();

        final response = functions.mdbDel(
          transaction,
          database.value,
          MDBvalue.fromUint8List(serialize(key)).addressOf,
          valuePointer,
        );
        if (response == KEY_NOT_EXISTS) {
          print('Key [$key] do not exists and can not be deleted');
          return;
        }
        ErrorHandler(response);
      },
    );
  }

  List<Object> findAll() {
    if (isClosed) {
      return null;
    }

    final boxes = <Object>[];

    _execute((functions, transaction, database) {
      final ffi.Pointer<ffi.Pointer<MDBcursor>> cursor = allocate();

      ErrorHandler(
        functions.mdbCursorOpen(transaction, database.value, cursor),
      );

      final ffi.Pointer<MDBvalue> key = allocate();
      final ffi.Pointer<MDBvalue> value = allocate();

      if (functions.mdbCursorGet(cursor.value, key, value, OP_MDB_FIRST) == 0) {
        boxes.add(_localize(value));

        while (functions.mdbCursorGet(cursor.value, key, value, OP_MDB_NEXT) ==
            0) {
          boxes.add(_localize(value));
        }
      }

      functions.mdbCursorClose(cursor.value);
    }, isReadOnly: true);

    return boxes;
  }

  bool get isClosed {
    if (_isClosed) {
      throw 'You cannot send events to a closed box';
    } else {
      return false;
    }
  }

  void close() {
    if (_isClosed) return;
    final functions = LMDBFunctions();

    ErrorHandler(functions.mdbEnvSync(_mdbEnv, 1));

    LMDBFunctions().mdbEnvClose(_mdbEnv);

    _boxes.remove(_path);

    _isClosed = true;

    print("box at $_path is closed.");
  }

  bool isOpen() => !_isClosed;

  void _execute(
    void Function(LMDBFunctions, ffi.Pointer<MDBtxn>, ffi.Pointer<ffi.IntPtr>)
        block, {
    bool isReadOnly = false,
  }) {
    final functions = LMDBFunctions();

    final pointerToTransaction = allocate<ffi.Pointer<MDBtxn>>();

    ErrorHandler(
      functions.mdbTxnBegin(
        _mdbEnv,
        ffi.nullptr,
        isReadOnly ? MDB_RDONLY : 0,
        pointerToTransaction,
      ),
    );

    final transaction = pointerToTransaction.value;

    if (_mdbDbi?.value == null) {
      _mdbDbi = allocate();

      ErrorHandler(
        functions.mdbDbiOpen(
          transaction,
          Utf8.toUtf8("data"),
          MDB_CREATE,
          _mdbDbi,
        ),
      );
    }

    block(functions, transaction, _mdbDbi);

    ErrorHandler(functions.mdbTxnCommit(transaction));
  }

  Object _localize(ffi.Pointer<MDBvalue> pointer) {
    final dataInBytes =
        pointer.ref.data.cast<ffi.Uint8>().asTypedList(pointer.ref.size);
    return deserialize(dataInBytes);
  }
}
